"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const bluebird_1 = require("bluebird");
const fs_extra_1 = require("fs-extra");
const lutils_1 = require("lutils");
const _1 = require("./");
const resolveDupes_1 = require("./lib/resolveDupes");
const schemaModuleMap_1 = require("./schemaModuleMap");
const typesToSchemas_1 = require("./typesToSchemas");
/**
 * An friendlier API for building multiple schemas by sharing configurations
 */
class TypeSchemaBuilder {
    constructor({ save = {}, compile = {}, reuseProgram = true, replaceWithRefs = true, emitErrors = true, deDupe = true, }) {
        this.compiled = [];
        this.builderConfigs = [];
        Object.assign(this, {
            saveConfig: save, compileConfig: compile,
            replaceWithRefs, emitErrors,
            deDupe,
        });
        const { fromFiles } = this.compileConfig;
        if (reuseProgram && fromFiles) {
            this.compileConfig.fromFiles = typesToSchemas_1.getTsProgram(fromFiles);
        }
    }
    //
    // Fluid interfaces
    //
    /** Add a IBuilderSchemaConfig to compile. Overrides any default values */
    add(configs) {
        if (!lutils_1.isArray(configs)) {
            configs = [configs];
        }
        this.builderConfigs.push(...configs);
        return this;
    }
    /** A simplified method, like #add(), to map a type to a name */
    addType(configs) {
        if (!lutils_1.isArray(configs)) {
            configs = [configs];
        }
        this.add(configs.map(({ type, name = type, id, optional = false, compile = {}, save = {} }) => {
            return lutils_1.merge({
                compile: { types: [{ type, name, id, optional }] },
                save: { name },
            }, { compile, save });
        }));
        return this;
    }
    //
    // Async interfaces
    //
    compileAndSave() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.compile();
            yield this.save();
            return this;
        });
    }
    compile() {
        return __awaiter(this, void 0, void 0, function* () {
            const refOverrides = {};
            if (this.replaceWithRefs) {
                const configs = [this, ...this.builderConfigs];
                for (const { compile = {} } of configs) {
                    const config = lutils_1.merge(lutils_1.clone(this.compileConfig || {}), { replaceWithRefs: true }, compile);
                    _1.extractRefsFromConfig(config).forEach(({ type, $ref }) => {
                        refOverrides[type] = $ref;
                    });
                }
            }
            /**
             * TODO: We can support ID annotation auto-parsing:
             * - Option 1: Read symbols without compiling, should save on time
             * - Option 2: Recurse through generated schemas and replace refs
             */
            yield bluebird_1.map(this.builderConfigs, (config) => __awaiter(this, void 0, void 0, function* () {
                const mergedConfig = lutils_1.merge(lutils_1.clone(this.compileConfig || {}), { refOverrides }, config.compile || {});
                const { errors, schemas } = yield typesToSchemas_1.typesToSchemas(mergedConfig);
                if (this.emitErrors && errors) {
                    errors.forEach(console.error);
                }
                this.compiled.push({
                    schemas,
                    errors,
                    config,
                });
            }));
            if (this.deDupe) {
                const countList = [];
                let jsonSchemas = this.compiled
                    .reduce((prev, { schemas: schemaItems }) => {
                    countList.push(schemaItems.length);
                    return [...prev, ...schemaItems.map(({ schema }) => schema)];
                }, []);
                jsonSchemas = resolveDupes_1.resolveDupes(jsonSchemas);
                this.compiled = this.compiled.map((compiled, index) => {
                    const ownedSchemas = jsonSchemas.splice(0, countList[index]);
                    return Object.assign({}, compiled, { schemas: compiled.schemas.map((item) => {
                            return Object.assign({}, item, { schema: ownedSchemas.shift() });
                        }) });
                });
            }
            return this.compiled;
        });
    }
    save(compiled = this.compiled) {
        return __awaiter(this, void 0, void 0, function* () {
            yield bluebird_1.map(compiled, ({ config, schemas: schemaConfigs, errors }) => __awaiter(this, void 0, void 0, function* () {
                if (errors) {
                    return;
                }
                const mergedConfig = lutils_1.merge(lutils_1.clone(this.saveConfig || {}), config.save || {});
                /** Whether to compile each seperately or together */
                const method = mergedConfig.asDefaultExport || mergedConfig.format === 'json'
                    ? 'seperate'
                    : 'together';
                if (method === 'seperate') {
                    return bluebird_1.map(schemaConfigs, (schemaConfig) => {
                        return typesToSchemas_1.saveSchemas(Object.assign({}, mergedConfig, { schemas: [schemaConfig] }));
                    });
                }
                else {
                    return typesToSchemas_1.saveSchemas(Object.assign({}, mergedConfig, { schemas: schemaConfigs }));
                }
            }));
        });
    }
    saveSchemaModuleMap({ schemas, moduleBase, fileName = '_map' }) {
        return __awaiter(this, void 0, void 0, function* () {
            const moduleMap = schemaModuleMap_1.createSchemaModuleMap({ schemas, moduleBase });
            const { directory } = this.saveConfig;
            if (!directory) {
                throw new Error('Cannot save the schemaModuleMap as no `save.directory` was specified');
            }
            const filePath = `${directory}/${fileName}.ts`;
            yield fs_extra_1.writeFile(filePath, schemaModuleMap_1.renderSchemaModuleMapToTs(moduleMap));
        });
    }
}
exports.TypeSchemaBuilder = TypeSchemaBuilder;
//# sourceMappingURL=TypeSchemaBuilder.js.map