"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const jst_1 = require("@jdw/jst");
const fs_extra_1 = require("fs-extra");
const lutils_1 = require("lutils");
const path_1 = require("path");
const typescript_json_schema_1 = require("typescript-json-schema");
const schemaRenderers_1 = require("./lib/schemaRenderers");
exports.defaultOptions = {
    required: true,
    noExtraProps: true,
    ignoreErrors: true,
    aliasRef: true,
};
function getTsProgram(fromFiles) {
    const program = lutils_1.isArray(fromFiles)
        ? typescript_json_schema_1.getProgramFromFiles(fromFiles)
        : fromFiles;
    return program;
}
exports.getTsProgram = getTsProgram;
/**
 * Reads TypeScript files using `typescript-json-schema` and returns both
 * errors and the resulting schemas.
 */
function typesToSchemas(config) {
    return __awaiter(this, void 0, void 0, function* () {
        const { types, dereference: doDereference = false, generator: inputGenerator, } = config;
        const schemas = [];
        const errors = [];
        const generator = inputGenerator || createTypeToSchemaGenerator(config);
        if (!generator) {
            throw new Error('Failure to create a generator for in `typesToSchemas`');
        }
        for (const { name, type, id, optional } of types) {
            let schema;
            try {
                schema = generator.getSchemaForSymbol(type);
            }
            catch (error) {
                if (optional) {
                    continue;
                }
                errors.push(error);
            }
            if (schema) {
                if (id && !schema.$id) {
                    schema.$id = id;
                }
                schema = removeUnusedJsonSchemaDefinitions(schema);
                if (doDereference) {
                    schema = jst_1.dereference(schema, undefined);
                }
                schemas.push({ name, type, schema });
            }
        }
        return {
            errors: errors.length ? errors : undefined,
            schemas,
        };
    });
}
exports.typesToSchemas = typesToSchemas;
function extractRefsFromConfig({ refOverrides, replaceWithRefs, types }) {
    const refsToReplace = [];
    // Add in all refOverrides to replace them with $refs
    if (refOverrides) {
        for (const type of Object.keys(refOverrides)) {
            refsToReplace.push({ type, $ref: refOverrides[type] });
        }
    }
    // When a `type` def has an associated id, we can use that as a $ref
    if (replaceWithRefs && types) {
        for (const { type, id: $ref } of types) {
            if (!$ref) {
                continue;
            }
            refsToReplace.push({ type, $ref });
        }
    }
    return refsToReplace;
}
exports.extractRefsFromConfig = extractRefsFromConfig;
function findRefsInSchema(schema, refs) {
    const matches = [];
    if (!(schema instanceof Object)) {
        return matches;
    }
    for (const key of Object.keys(schema)) {
        if (key === '$ref' && refs.includes(schema[key])) {
            matches.push(schema[key]);
            continue;
        }
        matches.push(...findRefsInSchema(schema[key], refs));
    }
    return matches;
}
function removeUnusedJsonSchemaDefinitions(schema) {
    if (!schema.definitions) {
        return schema;
    }
    const newSchema = Object.assign({}, schema, { definitions: Object.assign({}, schema.definitions) });
    const definitionKeys = Object.keys(schema.definitions).map((key) => `#/definitions/${key}`);
    const definitionsInUse = findRefsInSchema(schema, definitionKeys);
    Object.keys(newSchema.definitions).forEach((key) => {
        if (!definitionsInUse.includes(`#/definitions/${key}`)) {
            delete newSchema.definitions[key];
        }
    });
    return newSchema;
}
exports.removeUnusedJsonSchemaDefinitions = removeUnusedJsonSchemaDefinitions;
function createTypeToSchemaGenerator(config) {
    const { fromFiles, options } = config;
    const program = getTsProgram(fromFiles);
    const generator = typescript_json_schema_1.buildGenerator(program, Object.assign({}, exports.defaultOptions, options));
    const refsToReplace = extractRefsFromConfig(config);
    // Set the overrides for $refs in the generator
    for (const { type, $ref } of refsToReplace) {
        generator.setSchemaOverride(type, { $ref });
    }
    return generator;
}
exports.createTypeToSchemaGenerator = createTypeToSchemaGenerator;
/**
 * Saves schemas as multiple format, to a `.ts` or `.json` file, based on provided options.
 */
function saveSchemas({ schemas, directory, name, format, asDefaultExport = false }) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!schemas.length) {
            return;
        }
        if (asDefaultExport && schemas.length > 1) {
            throw new Error(`You are trying to default export more than one schema`);
        }
        yield fs_extra_1.mkdirs(directory);
        const filePath = path_1.join(directory, `${name}.${format}`);
        let file;
        if (format === 'ts') {
            file = schemaRenderers_1.renderSchemasToTs(schemas, { asDefaultExport });
        }
        if (format === 'json') {
            file = schemaRenderers_1.renderSchemasToJson(schemas);
        }
        if (!file) {
            throw new Error(`Invalid render format: ${format}`);
        }
        yield fs_extra_1.writeFile(filePath, file);
    });
}
exports.saveSchemas = saveSchemas;
/**
 * Creates an index file wiring up imports and exports
 */
function saveExports({ exports, directory, name, getImportPath, getImportPattern }) {
    return __awaiter(this, void 0, void 0, function* () {
        yield fs_extra_1.mkdirs(directory);
        const file = schemaRenderers_1.renderExportsToTs(exports, { getImportPath, getImportPattern });
        const filePath = path_1.join(directory, `${name}.ts`);
        yield fs_extra_1.writeFile(filePath, file);
    });
}
exports.saveExports = saveExports;
//# sourceMappingURL=typesToSchemas.js.map